#[cfg(feature = "Win32_Storage_Cabinets")]
#[doc = "Required features: `\"Win32_Storage_Cabinets\"`"]
pub mod Cabinets;
#[cfg(feature = "Win32_Storage_CloudFilters")]
#[doc = "Required features: `\"Win32_Storage_CloudFilters\"`"]
pub mod CloudFilters;
#[cfg(feature = "Win32_Storage_Compression")]
#[doc = "Required features: `\"Win32_Storage_Compression\"`"]
pub mod Compression;
#[cfg(feature = "Win32_Storage_DistributedFileSystem")]
#[doc = "Required features: `\"Win32_Storage_DistributedFileSystem\"`"]
pub mod DistributedFileSystem;
#[cfg(feature = "Win32_Storage_FileHistory")]
#[doc = "Required features: `\"Win32_Storage_FileHistory\"`"]
pub mod FileHistory;
#[cfg(feature = "Win32_Storage_FileSystem")]
#[doc = "Required features: `\"Win32_Storage_FileSystem\"`"]
pub mod FileSystem;
#[cfg(feature = "Win32_Storage_Imapi")]
#[doc = "Required features: `\"Win32_Storage_Imapi\"`"]
pub mod Imapi;
#[cfg(feature = "Win32_Storage_IndexServer")]
#[doc = "Required features: `\"Win32_Storage_IndexServer\"`"]
pub mod IndexServer;
#[cfg(feature = "Win32_Storage_InstallableFileSystems")]
#[doc = "Required features: `\"Win32_Storage_InstallableFileSystems\"`"]
pub mod InstallableFileSystems;
#[cfg(feature = "Win32_Storage_IscsiDisc")]
#[doc = "Required features: `\"Win32_Storage_IscsiDisc\"`"]
pub mod IscsiDisc;
#[cfg(feature = "Win32_Storage_Jet")]
#[doc = "Required features: `\"Win32_Storage_Jet\"`"]
pub mod Jet;
#[cfg(feature = "Win32_Storage_Nvme")]
#[doc = "Required features: `\"Win32_Storage_Nvme\"`"]
pub mod Nvme;
#[cfg(feature = "Win32_Storage_OfflineFiles")]
#[doc = "Required features: `\"Win32_Storage_OfflineFiles\"`"]
pub mod OfflineFiles;
#[cfg(feature = "Win32_Storage_OperationRecorder")]
#[doc = "Required features: `\"Win32_Storage_OperationRecorder\"`"]
pub mod OperationRecorder;
#[cfg(feature = "Win32_Storage_Packaging")]
#[doc = "Required features: `\"Win32_Storage_Packaging\"`"]
pub mod Packaging;
#[cfg(feature = "Win32_Storage_ProjectedFileSystem")]
#[doc = "Required features: `\"Win32_Storage_ProjectedFileSystem\"`"]
pub mod ProjectedFileSystem;
#[cfg(feature = "Win32_Storage_StructuredStorage")]
#[doc = "Required features: `\"Win32_Storage_StructuredStorage\"`"]
pub mod StructuredStorage;
#[cfg(feature = "Win32_Storage_Vhd")]
#[doc = "Required features: `\"Win32_Storage_Vhd\"`"]
pub mod Vhd;
#[cfg(feature = "Win32_Storage_Xps")]
#[doc = "Required features: `\"Win32_Storage_Xps\"`"]
pub mod Xps;
