#if 0
# This file is generated by cbindgen. DO NOT EDIT
#endif


extern "C" {

void root();

} // extern "C"

#if 0
# This is a simple test to ensure that trailers do not cause extra newlines in files
#endif
